from collections import namedtuple
from typing import TypeVar
import polars as pl
DataFrameType = TypeVar('DataFrameType', pl.LazyFrame, pl.DataFrame)

#import module to display tracebacks

IntegerRange = namedtuple("IntegerRange", ["datatype", "min_value", "max_value","bit_resolution"])

def calculate_int_range(datatype : pl.DataType )-> IntegerRange:
    """Function uses the name of the datatype to determine if it is signed or 
    unsigned and bit resolution and from that its range"""

    name = str(datatype)
    bit_resolution = int(name.lower().rsplit("int", maxsplit=1)[-1])                   
    unsigned = name[0].lower() == "u"
    if unsigned:
        min_value = 0
        max_value = 2**(bit_resolution)-1
    else:
        min_value = -2**(bit_resolution-1) 
        max_value = 2**(bit_resolution-1)-1
    return IntegerRange(datatype, min_value, max_value, bit_resolution)

TYPE_RANGES = frozenset({calculate_int_range(type) for type in pl.INTEGER_DTYPES})

def downcast_integers(frame: DataFrameType) -> DataFrameType:
    """ Function takes a polars frame and downcasts
    the integer columns to the smallest 
    possible datatype"""
     

    start_frame_type = type(frame)
    frame = frame.lazy()
    expressions = []
    
    for col, dtype in frame.schema.items():
        if dtype not in pl.INTEGER_DTYPES:
            continue
        #Get the min and max values of the column
        #LazyFrame->shortLazyFrame->shortDataFrame - > value at 0,0
        min_value = frame.select(col).min().collect().item()
        max_value = frame.select(col).max().collect().item()
        if max_value is None:
            continue
        if min_value is None:
            min_value = 0
        compatable_types = (type_range for type_range in TYPE_RANGES if type_range.min_value <= min_value and type_range.max_value >= max_value)
        sorted_types = sorted(compatable_types, key=lambda x: x.bit_resolution)
        if not sorted_types:
            continue
        selected_type = sorted_types[0]
        if selected_type.datatype == dtype:
            continue
        expressions.append(pl.col(col).cast(selected_type.datatype))
    frame = frame.with_columns(expressions) if expressions else frame
    frame = frame.collect() if start_frame_type == pl.DataFrame else frame
    return frame

def test_downcast_integers(datatype, mode= 'eager'):
    import numpy as np

    if datatype not in pl.INTEGER_DTYPES:
        return True

    #calculate test range
    type_range = calculate_int_range(datatype)
    #generate random 100 test points using Uint64 as part of the numpy function for unsigned datatypes
    test_data = np.random.randint(type_range.min_value, type_range.max_value, 100, 
               dtype=np.uint64 if type_range.min_value == 0 else np.int64)
    
    if min(test_data) ==0:
        frame = pl.DataFrame({"test": list(test_data)})
    else:
        frame = pl.DataFrame({"test": list(test_data)})
    null_frame = pl.DataFrame({'test': [None]}).with_columns(pl.col('test').cast(int))
    frame = pl.concat([frame, null_frame], how='vertical_relaxed')
    if mode=='lazy':
        frame = frame.lazy()
 
    #test if downcasted correctly
    downcasted_frame = downcast_integers(frame)
    downcasted_types = downcasted_frame.schema
    
    assert isinstance(downcasted_frame,type(frame)), f"Downcasted to {type(downcasted_frame)} but should be {type(frame)}"
    assert downcasted_types["test"] == datatype, f"Downcasted to {downcasted_types['test']} but should be {datatype}"
    

def test_downcast_integers_all():
    for dtype in pl.INTEGER_DTYPES:
        if 'UInt64' in str(dtype):
            print("Skipping UInt64")
            continue
        for mode in ['lazy', 'eager']:
            test_downcast_integers(dtype, mode=mode)
    print("All tests passed")

#test_downcast_integers_all()


from typing import TypeVar
import polars as pl
DataFrameType = TypeVar('DataFrameType', pl.LazyFrame, pl.DataFrame)


def downcast_strings(data_frame: DataFrameType) -> DataFrameType:
    original_schema = data_frame.schema


    transform_cols = [col for col in data_frame.columns if original_schema[col] == pl.Utf8]
    if len(transform_cols) == 0:
        return data_frame
    
    transformed_data_frame = data_frame.with_columns([pl.col(col).cast(pl.Categorical) for col in data_frame.columns if original_schema[col] == pl.Utf8])

    return transformed_data_frame
            

def test_downcast_strings_eager():
    data_frame = pl.DataFrame({
        "a": ["a", "b", "c"],
        "b": [1, 2, 3],
        "c": [True, False, True]
    })

    assert data_frame.schema["a"] == pl.Utf8
    assert data_frame.schema["b"] == pl.Int64
    assert data_frame.schema["c"] == pl.Boolean

    data_frame = downcast_strings(data_frame)

    assert data_frame.schema["a"] == pl.Categorical, f"Expected {data_frame.schema['a']} to be {pl.Categorical}"
    assert data_frame.schema["b"] == pl.Int64,f"Expected {data_frame.schema['b']} to be {pl.Int64}"
    assert data_frame.schema["c"] == pl.Boolean,f"Expected {data_frame.schema['c']} to be {pl.Boolean}"

def test_downcast_strings_lazy():
    data_frame = pl.DataFrame({
        "a": ["a", "b", "c"],
        "b": [1, 2, 3],
        "c": [True, False, True]
    }).lazy()

    assert data_frame.schema["a"] == pl.Utf8
    assert data_frame.schema["b"] == pl.Int64
    assert data_frame.schema["c"] == pl.Boolean

    data_frame = downcast_strings(data_frame)

    data_frame = data_frame.collect()
    assert data_frame.schema["a"] == pl.Categorical, f"Expected {data_frame.schema['a']} to be {pl.Categorical}"
    assert data_frame.schema["b"] == pl.Int64,f"Expected {data_frame.schema['b']} to be {pl.Int64}"
    assert data_frame.schema["c"] == pl.Boolean,f"Expected {data_frame.schema['c']} to be {pl.Boolean}"



from typing import TypeVar
import polars as pl
DataFrameType = TypeVar('DataFrameType', pl.LazyFrame, pl.DataFrame)

from .integers import downcast_integers
from .strings import downcast_strings

def compress_frame(frame: DataFrameType) -> DataFrameType:
    categorized = downcast_strings(frame)
    integer_bit_shortened = downcast_integers(categorized)
    return integer_bit_shortened

def real_test_compresss_file():
    data_frame = pl.scan_csv("exampledata1.csv", infer_schema_length=1)
    print(f"Original schema: {data_frame.schema}")
    data_frame = compress_frame(data_frame).collect()
    print(f"Schema: {data_frame.schema}")



from typing import TypeVar, Iterable
import polars as pl
DataFrameType = TypeVar('DataFrameType', pl.LazyFrame, pl.DataFrame)


def add_missing_columns(frames: Iterable[DataFrameType]) -> list[DataFrameType]:
    """
    Add columns that are missing from a frame
    """
    columns = set()
    frames = []
    for frame in frames:
        columns.update(frame.columns)
    for frame in frames:
        missing_columns = columns - set(frame.columns)
        frames.append(frame.with_columns([pl.lit(None).alias(column) for column in missing_columns]))
    return frames


def majority_schema_concat(frames: Iterable[DataFrameType]) -> DataFrameType:
    """
    Concatenate a list of frames into a single frame
    """
    extra_column_frame = add_missing_columns(frames)
    return pl.concat(extra_column_frame, how="vertical_relaxed")
        


import polars as pl
from typing import TypeVar, Generator, Sequence, Set

def recursive_pivot_iter(frame: pl.DataFrame,
        values: str,
        index: str | Sequence[str] | None,
        columns: str,
        maintain_order: bool = True,
        sort_columns: bool = False,
        seperator: str = "_")-> Generator[pl.DataFrame, None, None]:

    current_frame = frame
    while True:
        new_frame_cols = current_frame.select(columns).unique().to_series().to_list()
        new_frame = current_frame.pivot(values=values,index=index,columns=columns, maintain_order=maintain_order,sort_columns=sort_columns,separator=seperator, aggregate_function='min')
        if 0 in new_frame.shape:
            break
        yield new_frame
        print(new_frame.columns)
        max_value_frame = new_frame.select(new_frame_cols).max()
        max_stamp_overall = max_value_frame.transpose().max().item()
        current_frame = current_frame.filter(pl.col(values) > max_stamp_overall)

def recursive_pivot_concat(frame: pl.DataFrame,
        values: str,
        index: str | Sequence[str] | None,
        columns: str,
        maintain_order: bool = True,
        sort_columns: bool = False,
        separator: str = "_") -> pl.DataFrame:
        unique_values = list(map(str,frame.select(columns).unique(maintain_order=True).to_series().to_list()))
        frames = list(recursive_pivot_iter(frame, values, index, columns, maintain_order, sort_columns, separator))
        all_cols_ordered = [col for col in frame.columns if col != values and col!=columns] + unique_values
        frames_with_missing = [frame.with_columns([pl.lit(None).alias(col) for col in all_cols_ordered if col not in frame.columns]).select(all_cols_ordered) for frame in frames]
        return pl.concat(frames_with_missing, how='vertical_relaxed')



# Test for recursive_pivot_iter and recursive_pivot_concat
def test_recursive_pivot():
    # Initialize DataFrame with timestamps and garages
    df = pl.DataFrame({
        "Timestamp": [1, 2, 3, 4, 5, 6, 7],
        "Garage": ["A", "A", "B", "A", "B", "C", "C"],
        "Value": [10, 10, 10, 10, 10, 10, 10]
    })

    # Use recursive_pivot_iter function
    result = recursive_pivot_concat(df, index="Value", columns="Garage", values="Timestamp")
    # There should be 3 unique DataFrames since each garage has different min timestamps
    """ 
    ┌───────┬──────┬──────┬─────┐
    │ Value ┆ A    ┆ B    ┆ C   │
    │ ---   ┆ ---  ┆ ---  ┆ --- │
    │ i64   ┆ i64  ┆ i64  ┆ i64 │
    ╞═══════╪══════╪══════╪═════╡
    │ 10    ┆ 1    ┆ 3    ┆ 6   │
    │ 10    ┆ null ┆ null ┆ 7   │
    └───────┴──────┴──────┴─────┘"""
    print(result)
    expected_frame = pl.DataFrame({
        "Value": [10, 10],
        "A": [1, None],
        "B": [3, None],
        "C": [6, 7]})
       
    assert result.frame_equal(expected_frame)


# Run the test
#test_recursive_pivot()
